import asyncio
import os

from unittest import TestCase
from wafl.config import Configuration
from wafl.connectors.prompt_template import PromptCreator
from wafl.connectors.remote.remote_llm_connector import RemoteLLMConnector

_path = os.path.dirname(__file__)


class TestConnection(TestCase):
    def test__connection_to_generative_model_can_generate_text(self):
        config = Configuration.load_local_config()
        connector = RemoteLLMConnector(config.get_value("llm_model"))
        prediction = asyncio.run(
            connector.predict(
                PromptCreator.create_from_one_instruction(
                    'Generate a full paragraph based on this chapter title "The first contact".'
                    "The theme of the paragraph is space opera. "
                )
            )
        )
        assert len(prediction) > 0

    def test__connection_to_generative_model_can_generate_text_within_tags(self):
        config = Configuration.load_local_config()
        connector = RemoteLLMConnector(config.get_value("llm_model"))
        connector._num_prediction_tokens = 200
        text = 'Generate a full paragraph based on this chapter title " The First Contact". The theme of the paragraph is space opera. Include the characters "Alberto" and "Maria". Write at least three sentences.'
        prompt = f"""
<task>
Complete the following task and add <|EOS|> at the end: {text}
</task>
<result>
                """.strip()

        prediction = asyncio.run(
            connector.predict(PromptCreator.create_from_one_instruction(prompt))
        )
        print(prediction)
        assert len(prediction) > 0

    def test__connection_to_generative_model_can_generate_a_python_list(self):
        config = Configuration.load_local_config()
        connector = RemoteLLMConnector(config.get_value("llm_model"))
        connector._num_prediction_tokens = 200
        prompt = "Generate a Python list of 4 chapters names for a space opera book. The output needs to be a python list of strings: "
        prediction = asyncio.run(
            connector.predict(PromptCreator.create_from_one_instruction(prompt))
        )
        assert len(prediction) > 0



from dataclasses import dataclass
from datetime import datetime
from typing import List


@dataclass
class Utterance:
    text: str
    speaker: str
    timestamp: float

    def __init__(self, text: str, speaker: str, timestamp: float = None):
        self.text = text
        self.speaker = speaker
        self.timestamp = timestamp
        if self.timestamp is None:
            self.timestamp = datetime.now().timestamp()

    def to_dict(self):
        return {"text": self.text, "speaker": self.speaker, "timestamp": self.timestamp}

    def __str__(self):
        return f"{self.speaker}: {self.text}"


@dataclass
class Conversation:
    utterances: List[Utterance] = None

    def __init__(self, utterances: List[Utterance] = None):
        self.utterances = utterances

    def add_utterance(self, utterance: Utterance):
        if self.utterances is None:
            self.utterances = []

        if (
            len(self.utterances) > 0
            and utterance.text == self.utterances[-1].text
            and utterance.speaker == self.utterances[-1].speaker
        ):
            return

        if (
            len(self.utterances) > 0
            and utterance.speaker == self.utterances[-1].speaker
        ):
            self.utterances[-1].text += "\n" + utterance.text
            return

        self.utterances.append(utterance)
        self.utterances = sorted(self.utterances, key=lambda x: x.timestamp)

    def insert_utterance(self, new_utterance: Utterance):
        """
        Insert a new utterance into the conversation at the timestamp defined in the utterance.
        :param new_utterance:
        """
        if self.utterances is None:
            self.utterances = []

        new_utterances = []
        already_inserted = False
        for utterance in self.utterances:
            if (
                not already_inserted
                and utterance.speaker == new_utterance.speaker
                and utterance.timestamp == new_utterance.timestamp
            ):
                new_utterances.append(
                    Utterance(
                        new_utterance.text,
                        new_utterance.speaker,
                        new_utterance.timestamp,
                    )
                )
                already_inserted = True

            new_utterances.append(
                Utterance(utterance.text, utterance.speaker, utterance.timestamp)
            )

        self.utterances = new_utterances

    def get_last_n(self, n: int, stop_at_string: str = None) -> "Conversation":
        if not self.utterances:
            return Conversation()

        utterances_to_return: List[Utterance] = []
        stop_at_next = False
        for utterance in reversed(self.utterances):
            utterances_to_return.append(utterance)
            if stop_at_next:
                break
            if stop_at_string and stop_at_string in utterance.text:
                stop_at_next = True
                continue

            if len(utterances_to_return) == n:
                break
        utterances_to_return.reverse()
        return Conversation(utterances_to_return)

    def get_last_speaker_utterances(self, speaker: str, n: int) -> List[str]:
        if not self.utterances:
            return []

        return [
            utterance.text
            for utterance in self.utterances
            if utterance.speaker == speaker
        ][-n:]

    def get_first_timestamp(self) -> float:
        return self.utterances[0].timestamp if self.utterances else None

    def get_last_timestamp(self) -> float:
        return self.utterances[-1].timestamp if self.utterances else None

    def to_dict(self):
        return [utterance.to_dict() for utterance in self.utterances]

    def get_utterances_list(self) -> List[Utterance]:
        if not self.utterances:
            return []
        return self.utterances

    def __len__(self):
        if not self.utterances:
            return 0
        return len(self.utterances)



import random
import sys
import threading

from flask import render_template, redirect

from wafl.interface.list_interface import ListInterface
from wafl.interface.voice_interface import VoiceInterface
from wafl.scheduler.scheduler import Scheduler
from wafl.handlers.conversation_handler import ConversationHandler
from wafl.logger.local_file_logger import LocalFileLogger
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.queue_interface import QueueInterface
from wafl.config import Configuration
from wafl.runners.routes import get_app, add_new_routes
from wafl.handlers.web_handler import WebHandler

app = get_app()
_logger = LocalFileLogger()


@app.route("/create_new_instance", methods=["POST"])
def create_new_web_and_audio_instance():
    conversation_id = random.randint(0, sys.maxsize)
    result = create_web_and_audio_scheduler_and_webserver_loop(conversation_id)
    add_new_routes(conversation_id, result["web_server_handler"])
    thread = threading.Thread(target=result["scheduler"].run)
    thread.start()
    return redirect(f"{conversation_id}/index")


@app.route("/")
async def index_web_and_audio_():
    return render_template("selector.html")


def create_web_and_audio_scheduler_and_webserver_loop(conversation_id):
    config = Configuration.load_local_config()
    interface = ListInterface([VoiceInterface(config), QueueInterface()])
    interface.activate()
    conversation_events = ConversationEvents(
        config=config,
        interface=interface,
        logger=_logger,
    )
    conversation_loop = ConversationHandler(
        interface,
        conversation_events,
        _logger,
        activation_word=config.get_value("waking_up_word"),
    )
    web_handler = WebHandler(interface, config, conversation_id, conversation_events)
    return {
        "scheduler": Scheduler([conversation_loop, web_handler]),
        "web_server_handler": web_handler,
    }


def run_app():
    app.run(
        host="0.0.0.0",
        port=Configuration.load_local_config().get_value("frontend_port"),
    )


if __name__ == "__main__":
    run_app()



from typing import List

from wafl.dataclasses.facts import Fact


class BaseReader:
    def read(self, file_path: str) -> str:
        raise NotImplementedError()

    def get_chunks(self, filename: str) -> List[Fact]:
        raise NotImplementedError()

    def _chunk_text(self, text: str, size: int, overlap: int) -> List[str]:
        chunks = []
        for i in range(0, len(text), size - overlap):
            chunks.append(text[i : i + size])
        return chunks



import os
import textwrap

from wafl.connectors.factories.llm_connector_factory import LLMConnectorFactory
from wafl.connectors.prompt_template import PromptTemplate
from wafl.interface.conversation import Utterance, Conversation

_path = os.path.dirname(__file__)


class Entailer:
    def __init__(self, config):
        self._connector = LLMConnectorFactory.get_connector(config)
        self._config = config

    async def left_entails_right(self, lhs: str, rhs: str, dialogue) -> str:
        prompt = await self._get_answer_prompt(lhs, rhs, dialogue)
        result = await self._connector.generate(prompt)
        result = self._clean_result(result)
        return result == "yes"

    async def _get_answer_prompt(self, lhs, rhs, dialogue):
        return PromptTemplate(
            system_prompt="",
            conversation=self._get_dialogue_prompt(lhs, rhs, dialogue),
        )

    def _clean_result(self, result):
        result = result.replace("</task>", "")
        result = result.split("\n")[0]
        result = result.strip()
        return result.lower()

    def _get_dialogue_prompt(self, dialogue, lhs, rhs):
        text = f"""
Your task is to determine whether two sentences are similar.
1) {lhs.lower()}
2) {rhs.lower()}
Please answer "yes" if the two sentences are similar or "no" if not: 
        """.strip()
        return Conversation([Utterance(speaker="user", text=text)])



import os
import sys

from wafl.config import create_initial_files
from wafl.knowledge.indexing_implementation import add_to_index
from wafl.parsing.preprocess import remove_preprocessed
from wafl.run import (
    run_from_command_line,
    run_testcases,
    print_incipit,
    download_models,
)
from wafl.runners.run_from_actions import run_action


def print_help():
    print("\n")
    print("These are the available commands:")
    print("> wafl init: Initialize the current folder")
    print("> wafl add <PATH>: Add the file or folder at <PATH> to the index")
    print(
        "> wafl run: Starts the chatbot on the web interface and the audio interface."
    )
    print("> wafl run-cli: Run a cli version of the chatbot")
    print("> wafl run-audio: Run a voice-powered version of the chatbot")
    print("> wafl run-server: Run a webserver version of the chatbot")
    print("> wafl run-tests: Run the tests in testcases.txt")
    print("> wafl add <PATH>: Add the file or folder at <PATH> to the index")
    print(
        "> wafl run-action <ACTION_NAME>: Run the action <ACTION_NAME> from actions.yaml"
    )
    print("> wafl help: Show this help message")
    print()


def add_cwd_to_syspath():
    sys.path.append(os.getcwd())


def process_cli():
    add_cwd_to_syspath()
    print_incipit()

    arguments = sys.argv
    if len(arguments) > 1:
        command = arguments[1]

        if command == "init":
            create_initial_files()
            download_models()

        elif command == "run":
            from wafl.runners.run_web_and_audio_interface import run_app

            run_app()
            remove_preprocessed("/")

        elif command == "run-cli":
            run_from_command_line()
            remove_preprocessed("/")

        elif command == "run-audio":
            from wafl.runners.run_from_audio import run_from_audio

            run_from_audio()
            remove_preprocessed("/")

        elif command == "run-server":
            from wafl.runners.run_web_interface import run_server_only_app

            run_server_only_app()
            remove_preprocessed("/")

        elif command == "run-tests":
            run_testcases()
            remove_preprocessed("/")

        elif command == "run-action":
            if len(arguments) > 2:
                action_name = arguments[2]
                run_action(action_name)

            else:
                print("Please provide the action name as the second argument.")

        elif command == "add":
            if len(arguments) > 2:
                path = arguments[2]
                add_to_index(path)

            else:
                print("Please provide the path as the second argument.")

        elif command == "help":
            print_help()

        else:
            print("Unknown argument.\n")
    else:
        print_help()


def main():
    try:
        process_cli()

    except RuntimeError as e:
        print(e)
        print("WAFL ended due to the exception above.")



import os
import textwrap
from typing import List

from wafl.connectors.factories.llm_connector_factory import LLMConnectorFactory
from wafl.connectors.prompt_template import PromptTemplate

_path = os.path.dirname(__file__)


class InformationClient:
    def __init__(self, config):
        self._connector = LLMConnectorFactory.get_connector(config)
        self._config = config

    async def get_information(self) -> str:
        return await self._connector.get_information()



from wafl.handlers.web_interface_implementation import get_html_from_dialogue_item


class MessagesCreator:
    def __init__(self, interface):
        self._interface = interface
        self._toggled_windows = []

    def toggle_logs(self):
        if "logs" in self._toggled_windows:
            self._toggled_windows.remove("logs")
        else:
            self._toggled_windows.append("logs")

    async def get_messages_window(self):
        conversation = ""
        conversation += await self._get_dialogue()
        if "logs" in self._toggled_windows:
            conversation += await self._get_logs()

        return conversation

    async def _get_dialogue(self):
        dialogue_items = self._interface.get_utterances_list()
        dialogue_items = [
            get_html_from_dialogue_item(item) for item in dialogue_items[::-1]
        ]
        conversation = (
            "<div id='dialogue' class='dialogue overflow-y-scroll rounded-lg'>"
        )
        conversation += "".join(dialogue_items)
        return conversation

    async def _get_logs(self):
        choices = self._interface.get_choices_and_timestamp()
        choices = [
            (
                item[0],
                "<div class='log-row'>" + item[1] + "</div>",
            )
            for item in choices
        ]
        facts = self._interface.get_facts_and_timestamp()
        facts = [
            (
                item[0],
                "<div class='log-row'>" + item[1] + "</div>",
            )
            for item in facts
        ]

        choices_and_facts = choices + facts
        choices_and_facts = sorted(choices_and_facts, key=lambda x: x[0])[::-1]
        choices_and_facts = [item[1] for item in choices_and_facts]
        conversation = "</div>"
        conversation += "<div id='logs' class='logs shadow-lg overflow-y-scroll rounded-lg width' style='flex-direction: column-reverse;'>"
        conversation += "".join(choices_and_facts)
        conversation += "</div>"
        return conversation



import os
from unittest import TestCase

from wafl.connectors.prompt_template import PromptCreator
from wafl.interface.conversation import Conversation, Utterance

_path = os.path.dirname(__file__)


class TestPrompts(TestCase):
    def test_utterance(self):
        utterance = Utterance(
            text="Hello", speaker="user", timestamp="2022-01-01T00:00:00"
        )
        self.assertEqual(
            utterance.to_dict(),
            {"text": "Hello", "speaker": "user", "timestamp": "2022-01-01T00:00:00"},
        )

    def test_conversation(self):
        utterance1 = Utterance(text="Hello", speaker="user", timestamp=2)
        utterance2 = Utterance(text="Hi", speaker="bot", timestamp=1)
        conversation = Conversation(utterances=[utterance1, utterance2])
        self.assertEqual(
            conversation.to_dict(),
            [
                {
                    "text": "Hello",
                    "speaker": "user",
                    "timestamp": 2,
                },
                {
                    "text": "Hi",
                    "speaker": "bot",
                    "timestamp": 1,
                },
            ],
        )

    def test_prompt(self):
        utterance1 = Utterance(text="Hello", speaker="user", timestamp=2)
        utterance2 = Utterance(text="Hi", speaker="bot", timestamp=1)
        conversation = Conversation(utterances=[utterance1, utterance2])
        prompt = PromptCreator.create(system_prompt="Hello", conversation=conversation)
        self.assertEqual(
            prompt.to_dict(),
            {
                "system_prompt": "Hello",
                "conversation": [
                    {
                        "text": "Hello",
                        "speaker": "user",
                        "timestamp": 2,
                    },
                    {
                        "text": "Hi",
                        "speaker": "bot",
                        "timestamp": 1,
                    },
                ],
            },
        )



import json

import aiohttp
import asyncio

from wafl.connectors.base_llm_connector import BaseLLMConnector
from wafl.connectors.prompt_template import PromptTemplate
from wafl.variables import is_supported


class RemoteLLMConnector(BaseLLMConnector):
    _max_tries = 3
    _max_reply_length = 1024
    _num_prediction_tokens = 200
    _cache = {}

    def __init__(self, config, last_strings=None, num_replicas=1):
        super().__init__(last_strings)
        host = config["model_host"]
        port = config["model_port"]
        self._default_temperature = config["temperature"]
        self._server_url = f"https://{host}:{port}/predictions/bot"
        self._num_replicas = num_replicas

        try:
            loop = asyncio.get_running_loop()

        except RuntimeError:
            loop = None

        if (not loop or (loop and not loop.is_running())) and not asyncio.run(
            self.check_connection()
        ):
            raise RuntimeError("Cannot connect a running LLM.")

    async def predict(
        self,
        prompt: PromptTemplate,
        temperature=None,
        num_tokens=None,
        num_replicas=None,
    ) -> [str]:
        if not temperature:
            temperature = self._default_temperature

        if not num_tokens:
            num_tokens = self._num_prediction_tokens

        if not num_replicas:
            num_replicas = self._num_replicas

        payload = {
            "data": prompt.to_dict(),
            "temperature": temperature,
            "num_tokens": num_tokens,
            "num_replicas": num_replicas,
        }

        for _ in range(self._max_tries):
            async with aiohttp.ClientSession(
                conn_timeout=6000,
                connector=aiohttp.TCPConnector(ssl=False),
            ) as session:
                async with session.post(self._server_url, json=payload) as response:
                    answer = json.loads(await response.text())
                    status = answer["status"]
                    if status != "success":
                        raise RuntimeError(f"Error in prediction: {answer}")
                    return answer["prediction"].split("<||>")

        return [""]

    async def check_connection(self):
        payload = {
            "data": {
                "system_prompt": "Hello!",
                "conversation": [{"speaker": "user", "text": "Hi!"}],
            },
            "temperature": 0.6,
            "num_tokens": 1,
            "last_strings": self._important_strings,
            "num_replicas": self._num_replicas,
        }
        try:
            async with aiohttp.ClientSession(
                conn_timeout=3, connector=aiohttp.TCPConnector(ssl=False)
            ) as session:
                async with session.post(self._server_url, json=payload) as response:
                    answer = json.loads(await response.text())
                    wafl_llm_version = answer["version"]
                    print(f"Connected to wafl-llm v{wafl_llm_version}.")
                    if not is_supported(wafl_llm_version):
                        print("This version of wafl-llm is not supported.")
                        print("Please update wafl-llm.")
                        raise aiohttp.client.InvalidURL

                    return True

        except aiohttp.client.InvalidURL:
            print()
            print("Is the wafl-llm running?")
            print("Please run 'bash start-llm.sh' (see docs for explanation).")
            print()

        return False

    async def get_information(self):
        payload = {
            "data": {
                "system_prompt": "Hello!",
                "conversation": [{"speaker": "user", "text": "Hi!"}],
            },
            "temperature": 0.6,
            "num_tokens": 1,
            "last_strings": self._important_strings,
            "num_replicas": self._num_replicas,
        }
        async with aiohttp.ClientSession(
            conn_timeout=6000,
            connector=aiohttp.TCPConnector(ssl=False),
        ) as session:
            async with session.post(self._server_url, json=payload) as response:
                answer = json.loads(await response.text())
                status = answer["status"]
                if status != "success":
                    raise RuntimeError(f"Error in prediction: {answer}")
                return {
                    "model_name": answer["model"],
                    "backend_version": answer["version"],
                }



from dataclasses import dataclass

from wafl.interface.conversation import Conversation, Utterance


@dataclass
class PromptTemplate:
    system_prompt: str = ""
    conversation: "Conversation" = None

    def to_dict(self):
        return {
            "system_prompt": self.system_prompt,
            "conversation": self.conversation.to_dict() if self.conversation else [],
        }


class PromptCreator:
    @staticmethod
    def create(system_prompt: str, conversation: Conversation) -> PromptTemplate:
        prompt = PromptTemplate()
        prompt.system_prompt = system_prompt
        prompt.conversation = conversation
        return prompt

    @staticmethod
    def create_from_one_instruction(instruction: str) -> PromptTemplate:
        return PromptTemplate(
            system_prompt="",
            conversation=Conversation(
                utterances=[Utterance(speaker="user", text=instruction)]
            ),
        )



from dataclasses import dataclass
from typing import Union


@dataclass
class Fact:
    text: Union[str, dict]
    is_question: bool = False
    variable: str = None
    is_interruption: bool = False
    source: str = None
    destination: str = None
    metadata: Union[str, dict] = None

    def toJSON(self):
        return str(self)



import re

from wafl.simple_text_processing.normalize import normalized


def input_is_valid(text):
    if not text.strip():
        return False

    if normalized(text) != "no" and len(text) <= 2:
        return False

    return True


def remove_text_between_brackets(text: str) -> str:
    return re.sub(r"(\[.*?\])", "", text)


def remove_unclear(text: str) -> str:
    return text.replace("[unclear]", "").strip()



from importlib import import_module
from inspect import getmembers, isfunction
from typing import List, Tuple
from wafl.answerer.answerer_implementation import (
    substitute_memory_in_answer_and_get_memories_if_present,
    create_one_liner,
    get_text_from_facts_and_thresholds,
    add_dummy_utterances_to_continue_generation,
    add_memories_to_facts,
    execute_results_in_answer,
)
from wafl.answerer.base_answerer import BaseAnswerer
from wafl.answerer.rule_maker import RuleMaker
from wafl.connectors.clients.llm_chat_client import LLMChatClient
from wafl.dataclasses.dataclasses import Query, Answer
from wafl.interface.conversation import Conversation
from wafl.simple_text_processing.questions import is_question


class DialogueAnswerer(BaseAnswerer):
    def __init__(self, config, knowledge, interface, code_path, logger):
        self._threshold_for_facts = 0.85
        self._client = LLMChatClient(config)
        self._knowledge = knowledge
        self._logger = logger
        self._interface = interface
        self._max_num_past_utterances = 5
        self._max_num_past_utterances_for_facts = 5
        self._max_num_past_utterances_for_rules = 2
        self._prior_facts_with_timestamp = []
        self._init_python_module(code_path.replace(".py", ""))
        self._prior_rules = []
        self._max_predictions = 3
        self._rule_creator = RuleMaker(
            knowledge,
            config,
            interface,
            max_num_rules=1,
        )

    async def answer(self, query_text: str) -> Answer:
        if self._logger:
            self._logger.write(f"Dialogue Answerer: the query is {query_text}")
        query = Query.create_from_text("The user says: " + query_text)
        conversation = self._interface.get_utterances_list_with_timestamp().get_last_n(
            self._max_num_past_utterances
        )
        rules_text = await self._get_relevant_rules(conversation)
        if not conversation:
            conversation = create_one_liner(query_text)
        conversational_timestamp = len(conversation)
        facts = await self._get_relevant_facts(
            query,
            has_prior_rules=bool(rules_text),
            conversational_timestamp=conversational_timestamp,
        )

        final_answer_text = ""
        for _ in range(self._max_predictions):
            original_answer_text = await self._client.get_answer(
                text=facts,
                rules_text=rules_text,
                dialogue=conversation,
            )
            await self._interface.add_fact(f"The bot predicts: {original_answer_text}")
            answer_text, memories = await self._apply_substitutions(
                original_answer_text
            )

            final_answer_text += answer_text

            if not memories:
                break

            facts = add_memories_to_facts(facts, memories)
            add_dummy_utterances_to_continue_generation(conversation, answer_text)

        if self._logger:
            self._logger.write(
                f"Answer within dialogue: The answer is {final_answer_text}"
            )

        return Answer.create_from_text(final_answer_text)

    async def _get_relevant_facts(
        self, query: Query, has_prior_rules: bool, conversational_timestamp: int
    ) -> str:
        memory = "\n".join([item[0] for item in self._prior_facts_with_timestamp])
        self._prior_facts_with_timestamp = self._get_prior_facts_with_timestamp(
            conversational_timestamp
        )
        facts_and_thresholds = await self._knowledge.ask_for_facts_with_threshold(
            query, is_from_user=True, threshold=self._threshold_for_facts
        )
        if facts_and_thresholds:
            facts = get_text_from_facts_and_thresholds(facts_and_thresholds, memory)
            self._prior_facts_with_timestamp.extend(
                (item, conversational_timestamp) for item in facts
            )
            memory = "\n".join([item[0] for item in self._prior_facts_with_timestamp])
            await self._interface.add_fact(f"The bot remembers the facts:\n{memory}")

        else:
            if is_question(query.text) and not has_prior_rules:
                memory += (
                    f"\nThe answer to {query.text} is not in the knowledge base."
                    "The bot can answer the question while informing the user that the answer was not retrieved"
                )

        return memory

    async def _get_relevant_rules(self, conversation: Conversation) -> List[str]:
        rules = await self._rule_creator.create_from_query(conversation)
        for rule in rules:
            if rule not in self._prior_rules:
                self._prior_rules.insert(0, rule)
        self._prior_rules = self._prior_rules[: self._max_num_past_utterances_for_rules]
        return self._prior_rules

    def _init_python_module(self, module_name):
        self._module = import_module(module_name)
        self._functions = [item[0] for item in getmembers(self._module, isfunction)]

    async def _apply_substitutions(self, original_answer_text):
        return await substitute_memory_in_answer_and_get_memories_if_present(
            await execute_results_in_answer(
                original_answer_text,
                self._module,
                self._functions,
            )
        )

    def _get_prior_facts_with_timestamp(
        self, conversational_timestamp: int
    ) -> List[Tuple[str, int]]:
        return [
            item
            for item in self._prior_facts_with_timestamp
            if item[1]
            > conversational_timestamp - self._max_num_past_utterances_for_facts
        ]



import os

import joblib
import yaml

from wafl.config import Configuration
from wafl.knowledge.single_file_knowledge import SingleFileKnowledge
from wafl.readers.reader_factory import ReaderFactory


async def _add_indices_to_knowledge(knowledge, text):
    indices = yaml.safe_load(text)
    if "paths" not in indices or not indices["paths"]:
        return knowledge

    for path in indices["paths"]:
        for root, _, files in os.walk(path):
            for file in files:
                reader = ReaderFactory.get_reader(file)
                for chunk in reader.get_chunks(os.path.join(root, file)):
                    await knowledge.add_fact(chunk)

    return knowledge


async def load_knowledge(config, logger=None):
    if ".yaml" in config.get_value("rules") and not any(
        item in config.get_value("rules") for item in [" ", "\n"]
    ):
        with open(config.get_value("rules")) as file:
            rules_txt = file.read()

    else:
        rules_txt = config.get_value("rules")

    index_filename = config.get_value("index")
    if not os.path.exists(index_filename):
        raise RuntimeError(f"Index file {index_filename} does not exist.")
    with open(index_filename) as file:
        index_txt = file.read()

    cache_filename = config.get_value("cache_filename")
    if os.path.exists(cache_filename):
        knowledge = joblib.load(cache_filename)
        if knowledge.hash == hash(rules_txt) and os.path.getmtime(
            cache_filename
        ) > os.path.getmtime(index_filename):
            await knowledge.initialize_retrievers()
            return knowledge

    knowledge = SingleFileKnowledge(config, rules_txt, logger=logger)
    knowledge = await _add_indices_to_knowledge(knowledge, index_txt)
    joblib.dump(knowledge, config.get_value("cache_filename"))
    await knowledge.initialize_retrievers()
    return knowledge


def add_to_index(path):
    config = Configuration.load_local_config()
    index_filename = config.get_value("index")
    with open(index_filename) as file:
        indices = yaml.safe_load(file.read())
        if path in indices["paths"]:
            return

        indices["paths"].append(path)

    with open(index_filename, "w") as file:
        yaml.dump(indices, file)



from typing import List

from wafl.dataclasses.dataclasses import Query
from wafl.dataclasses.rules import Rule


class RuleMaker:
    def __init__(
        self,
        knowledge: "Knowledge",
        config: "BaseConfig",
        interface: "BaseInterface",
        max_num_rules: int,
        max_recursion: int = 3,
    ):
        self._knowledge = knowledge
        self._config = config
        self._interface = interface
        self._max_num_rules = max_num_rules
        if not config.get_value("max_recursion"):
            self._max_indentation = max_recursion
        else:
            self._max_indentation = config.get_value("max_recursion")

    async def create_from_query(self, conversation: "Conversation") -> List[str]:
        rules = await self._get_rules(conversation)
        rules_texts = []
        for rule in rules:
            rules_text = rule.get_string_using_template(
                "- {effect}:\n{clauses}\n" + rule.indent_str
            )
            rules_texts.append(rules_text)
            await self._interface.add_fact(f"The bot remembers the rule:\n{rules_text}")

        return rules_texts

    async def _get_rules(self, conversation: "Conversation") -> List[Rule]:
        utterances = conversation.get_last_speaker_utterances("user", 1)
        rules = await self._knowledge.ask_for_rule_backward(
            Query.create_from_list(utterances), threshold=0.9
        )

        utterances = conversation.get_last_speaker_utterances("user", 2)
        rules.extend(
            await self._knowledge.ask_for_rule_backward(
                Query.create_from_list(utterances), threshold=0.8
            )
        )

        return rules[: self._max_num_rules]



import logging
import re

from wafl.connectors.utils import select_best_answer

_system_logger = logging.getLogger(__file__)

model = None
tokenizer = None


class BaseLLMConnector:
    _max_tries = 3
    _max_reply_length = 500
    _num_prediction_tokens = 200
    _cache = {}

    def __init__(self, important_strings=None):
        if not important_strings:
            self._important_strings = [
                "\nuser",
                "\nbot",
                "<|EOS|>",
                "</remember>",
                "</execute>\n",
                "</s>",
            ]

        else:
            self._important_strings = important_strings

    async def predict(self, prompt: str) -> [str]:
        raise NotImplementedError

    async def generate(self, prompt: "PromptTemplate") -> str:
        if str(prompt.to_dict()) in self._cache:
            return self._cache[str(prompt.to_dict())]

        text = select_best_answer(await self.predict(prompt), self._important_strings)
        end_set = set()
        for item in self._important_strings:
            if "</remember>" in item or "</execute>" in item:
                continue

            end_set.add(text.find(item))

        if -1 in end_set:
            end_set.remove(-1)

        end = len(text)
        if end_set:
            end = min(end_set)

        candidate_answer = text[:end].strip()
        candidate_answer = re.sub(r"(.*)<\|.*\|>", r"\1", candidate_answer).strip()

        if str(prompt.to_dict()) not in self._cache:
            self._cache[str(prompt.to_dict())] = candidate_answer

        if not candidate_answer:
            candidate_answer = "unknown"

        return candidate_answer



from logging import getLogger
from typing import List

from wafl.dataclasses.facts import Fact
from wafl.readers.base_reader import BaseReader

_logger = getLogger(__name__)


class TextReader(BaseReader):
    def __init__(self, chunk_size: int, overlap: int):
        self.chunk_size = chunk_size
        self.overlap = overlap

    def get_chunks(self, filename: str) -> List[Fact]:
        _logger.info(f"Reading text file: {filename}")
        with open(filename, "r") as file:
            chunks = self._chunk_text(file.read(), self.chunk_size, self.overlap)
            return [
                Fact(
                    text=chunk,
                    metadata={"filename": filename, "chunk_number": i},
                )
                for i, chunk in enumerate(chunks)
            ]



import asyncio
from unittest import TestCase

from wafl.config import Configuration
from wafl.testcases import ConversationTestCases

_wafl_example = """
rules:
  - the user says "hello":
    - You must answer the user by writing "Hello. What is your name"
    
  - the user says their name:
    - You must answer the user by writing "nice to meet you, NAME_OF_THE_USER!"
""".strip()


_test_case_greetings = """

test the greetings work
  user: Hello
  bot: Hello there! What is your name
  user: Bob
  bot: Nice to meet you, bob!

! test the greetings uses the correct name
  user: Hello
  bot: Hello there! What is your name
  user: Bob
  bot: Nice to meet you, unknown!


""".strip()


class TestConversationalTestCases(TestCase):
    def test_conversation_testcase_run_all(self):
        config = Configuration.load_local_config()
        config.set_value("rules", _wafl_example)
        testcase = ConversationTestCases(config=config, text=_test_case_greetings)
        self.assertTrue(asyncio.run(testcase.run()))



import json
import requests

from datetime import datetime, timedelta

# These are the coordinate of central London. Please change at your convenience
from wafl.exceptions import CloseConversation

latitude = "51.5074"
longitude = "0.1272"

_db_filename = "db.json"

def close_conversation():
    raise CloseConversation()

def check_today_weather():
    today = datetime.now().strftime("%Y-%m-%d")
    return check_weather_lat_long(latitude, longitude, today)


def check_tomorrow_weather():
    today = datetime.now()
    tomorrow = (today + timedelta(days=1)).strftime("%Y-%m-%d")
    return check_weather_lat_long(latitude, longitude, tomorrow)


def check_weather_lat_long(latitude, longitude, day):
    secrets = json.load(open("secrets.json"))
    print("HERE!")
    result = requests.get(
        f"https://rgw.5878-e94b1c46.eu-gb.apiconnect.appdomain.cloud/metoffice/production/v0/forecasts/point/daily?excludeParameterMetadata=true&includeLocationName=true&latitude={latitude}&longitude={longitude}",
        headers={
            "X-IBM-Client-Id": secrets["key"],
            "X-IBM-Client-Secret": secrets["secret"],
        },
    )
    print("HERE2!")
    data = result.json()
    if "features" not in data:
        return "There is a connection error to the weather API. Please try later"

    to_say = ""
    for item in data["features"][0]["properties"]["timeSeries"]:
        if day in item["time"]:
            print("HEREEE!")
            to_say += f"The temperature should be between {int(item['dayLowerBoundMaxTemp'])} and {int(item['dayUpperBoundMaxTemp'])}.\n"
            if item["dayProbabilityOfPrecipitation"] != 0:
                to_say += f"The probability of rain is {item['dayProbabilityOfPrecipitation']} percent.\n"

            if item["dayProbabilityOfSnow"] != 0:
                to_say += f"The probability of snow is {item['dayProbabilityOfSnow']} percent.\n"

    print(to_say)
    return to_say


def get_time():
    return datetime.now().strftime("%H:%M")


def get_date():
    return datetime.now().strftime("%Y-%m-%d")


def get_day():
    return datetime.now().strftime("%A")



def add_to_shopping_list(list_of_items_to_add):
    db = json.load(open(_db_filename))
    for item in list_of_items_to_add:
        if item not in db["shopping_list"]:
            db["shopping_list"].append(item)

    json.dump(db, open(_db_filename, "w"))

    return "Item added"


def remove_shopping_list(list_of_items_to_remove):
    db = json.load(open(_db_filename))
    for item in list_of_items_to_remove:
        if item in db["shopping_list"]:
            db["shopping_list"].remove(item)

    json.dump(db, open(_db_filename, "w"))

    return "Item removed"


def get_shopping_list():
    db = json.load(open(_db_filename))
    if db["shopping_list"] == []:
        return "nothing"

    return ", ".join(db["shopping_list"])


def write_to_file(filename, text):
    with open(filename, "w") as file:
        file.write(text)

    return f"File {filename} saved"



import os

from flask import Flask, abort
from flask_cors import CORS

_path = os.path.dirname(__file__)
app = Flask(
    __name__,
    static_url_path="",
    static_folder=os.path.join(_path, "../frontend/"),
    template_folder=os.path.join(_path, "../frontend/"),
)
CORS(app)

_routes_dict = {}


def get_app():
    return app


@app.route("/<conversation_id>/index", methods=["GET"])
async def get_index(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["index"]()


@app.route("/<conversation_id>/reset_conversation", methods=["POST"])
async def reset_conversation(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["reset_conversation"]()


@app.route("/<conversation_id>/reload_rules", methods=["POST"])
async def reload_rules(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["reload_rules"]()


@app.route("/<conversation_id>/check_new_messages", methods=["POST"])
async def check_new_messages(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["check_new_messages"]()


@app.route("/<conversation_id>/load_messages", methods=["POST", "GET"])
async def load_messages(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["load_messages"]()


@app.route("/<conversation_id>/input", methods=["POST", "GET"])
async def input(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["input"]()


@app.route("/<conversation_id>/output", methods=["POST"])
async def output(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["output"]()


@app.route("/<conversation_id>/thumbs_up", methods=["POST"])
async def thumbs_up(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["thumbs_up"]()


@app.route("/<conversation_id>/thumbs_down", methods=["POST"])
async def thumbs_down(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["thumbs_down"]()


@app.route("/<conversation_id>/toggle_logs", methods=["POST"])
async def toggle_logs(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["toggle_logs"]()


@app.route("/<conversation_id>/get_info", methods=["POST"])
async def get_info(conversation_id: str):
    if conversation_id not in _routes_dict:
        abort(404)
    return await _routes_dict[conversation_id]["get_info"]()


def add_new_routes(conversation_id: str, web_server_handler: "WebHandler"):
    _routes_dict[str(conversation_id)] = {
        "index": web_server_handler.index,
        "reset_conversation": web_server_handler.reset_conversation,
        "reload_rules": web_server_handler.reload_rules,
        "check_new_messages": web_server_handler.check_for_new_messages,
        "load_messages": web_server_handler.load_messages,
        "input": web_server_handler.handle_input,
        "output": web_server_handler.handle_output,
        "thumbs_up": web_server_handler.thumbs_up,
        "thumbs_down": web_server_handler.thumbs_down,
        "toggle_logs": web_server_handler.toggle_logs,
        "get_info": web_server_handler.get_info,
    }



from wafl.config import Configuration
from wafl.runners.run_web_interface import app
from wafl.variables import get_variables

if __name__ == "__main__":
    print(get_variables())
    app.run(
        host="0.0.0.0",
        port=Configuration.load_local_config().get_value("frontend_port"),
    )



import asyncio
from typing import List

from wafl.interface.base_interface import BaseInterface


class ListInterface(BaseInterface):
    def __init__(self, interfaces_list: List[BaseInterface]):
        super().__init__()
        self._interfaces_list = interfaces_list
        self._synchronize_interfaces()

    async def output(self, text: str, silent: bool = False):
        await asyncio.wait(
            [interface.output(text, silent) for interface in self._interfaces_list],
            return_when=asyncio.ALL_COMPLETED,
        )

    async def input(self) -> str:
        tasks = [interface.input() for interface in self._interfaces_list]
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        return done.pop().result()

    async def insert_input(self, text: str):
        await asyncio.wait(
            [interface.insert_input(text) for interface in self._interfaces_list],
            return_when=asyncio.ALL_COMPLETED,
        )

    def bot_has_spoken(self, to_set: bool = None):
        if to_set == None:
            return all(
                interface.bot_has_spoken() for interface in self._interfaces_list
            )

        for interface in self._interfaces_list:
            interface.bot_has_spoken(to_set)

    def activate(self):
        for interface in self._interfaces_list:
            interface.activate()
        super().activate()

    def deactivate(self):
        for interface in self._interfaces_list:
            interface.deactivate()
        super().deactivate()
        self._synchronize_interfaces()

    def add_hotwords(self, hotwords):
        for interface in self._interfaces_list:
            interface.add_hotwords(hotwords)

    def _synchronize_interfaces(self):
        for interface in self._interfaces_list:
            interface._utterances = self._utterances



import asyncio
import os

from flask import render_template, request, jsonify

from wafl.config import Configuration
from wafl.connectors.clients.information_client import InformationClient
from wafl.interface.base_interface import BaseInterface
from wafl.logger.history_logger import HistoryLogger
from wafl.scheduler.messages_creator import MessagesCreator

_path = os.path.dirname(__file__)


class WebHandler:
    def __init__(
        self,
        interface: BaseInterface,
        config: Configuration,
        conversation_id: int,
        conversation_events: "ConversationEvents",
    ):
        self._interface = interface
        self._history_logger = HistoryLogger(self._interface)
        self._conversation_id = conversation_id
        self._conversation_events = conversation_events
        self._prior_dialogue_items = ""
        self._messages_creator = MessagesCreator(self._interface)
        self._information_client = InformationClient(config)

    async def index(self):
        return render_template("index.html", conversation_id=self._conversation_id)

    async def handle_input(self):
        query = request.form["query"]
        await self._interface.insert_input(query)
        return f"""
    <textarea id="query" type="text"
           class='shadow-lg w-full'
           placeholder="{query}"
           name="query"
           hx-post="/{self._conversation_id}/input"
           hx-swap="outerHTML"
           hx-target="#query"
           hx-trigger="keydown[!shiftKey&&keyCode==13]"
    ></textarea>
        """.strip()

    async def reset_conversation(self):
        self._interface.reset_history()
        self._interface.deactivate()
        self._interface.activate()
        await self._conversation_events.reload_knowledge()
        self._conversation_events.reset_discourse_memory()
        await self._interface.output("Hello. How may I help you?")
        conversation = await self._messages_creator.get_messages_window()
        return conversation

    async def reload_rules(self):
        async with asyncio.Lock():
            print("Not implemented yet")

        return ""

    async def check_for_new_messages(self):
        conversation = await self._messages_creator.get_messages_window()
        if conversation != self._prior_dialogue_items:
            self._prior_dialogue_items = conversation
            return f"""
            <div id="load_conversation" 
               hx-post="/{self._conversation_id}/load_messages"
               hx-swap="innerHTML"
               hx-target="#messages"
               hx-trigger="load"
            ></div>"""

        else:
            self._prior_dialogue_items = conversation
            return "<div id='load_conversation'></div>"

    async def load_messages(self):
        conversation = await self._messages_creator.get_messages_window()
        return conversation

    async def handle_output(self):
        if not self._interface.output_queue:
            return jsonify({"text": "", "silent": False})

        output = self._interface.output_queue.pop(0)
        return jsonify(output)

    async def thumbs_up(self):
        self._history_logger.write("thumbs_up")
        return jsonify("")

    async def thumbs_down(self):
        self._history_logger.write("thumbs_down")
        return jsonify("")

    async def toggle_logs(self):
        self._messages_creator.toggle_logs()
        return jsonify("")

    async def get_info(self):
        info = await self._information_client.get_information()
        is_clicked = request.form.get("clicked")
        is_clicked = "false" if is_clicked == "true" else "true"
        infobox = ""
        if is_clicked == "true":
            infobox = f"""
            <div class="bg-white shadow-lg rounded-lg p-4 absolute w-max left-20">
                <div style="color:black;font-size:12px;"><b>Model name:</b> {info['model_name']}</div>
                <div style="color:black;font-size:12px;"><b>Backend version:</b> {info['backend_version']}</div>
            </div>
            """
        return f"""
        <a title="Info"
           hx-post="/{self._conversation_id}/get_info"
           hx-vals='{{"clicked": "{is_clicked}"}}'
           hx-swap="outerHTML"
           class="flex items-center p-2 rounded-lg text-white hover:bg-gray-700 group">
           <svg fill="#FFFFFF" xmlns="http://www.w3.org/2000/svg" x="0px" y="0px" class="w-6 h-6" viewBox="0 0 50 50">
           <path d="M 25 2 C 12.309295 2 2 12.309295 2 25 C 2 37.690705 12.309295 48 25 48 C 37.690705 48 48 37.690705 48 25 C 48 12.309295 37.690705 2 25 2 z M 25 4 C 36.609824 4 46 13.390176 46 25 C 46 36.609824 36.609824 46 25 46 C 13.390176 46 4 36.609824 4 25 C 4 13.390176 13.390176 4 25 4 z M 25 11 A 3 3 0 0 0 22 14 A 3 3 0 0 0 25 17 A 3 3 0 0 0 28 14 A 3 3 0 0 0 25 11 z M 21 21 L 21 23 L 22 23 L 23 23 L 23 36 L 22 36 L 21 36 L 21 38 L 22 38 L 23 38 L 27 38 L 28 38 L 29 38 L 29 36 L 28 36 L 27 36 L 27 21 L 26 21 L 22 21 L 21 21 z"></path>
           </svg>
            {infobox}
        </a>
        """

    async def run(self):
        print(f"New web server instance {self._conversation_id} running!")
        return



import pymupdf

from logging import getLogger
from typing import List
from wafl.dataclasses.facts import Fact
from wafl.readers.base_reader import BaseReader

_logger = getLogger(__name__)


class PdfReader(BaseReader):
    def __init__(self, chunk_size: int, overlap: int):
        self.chunk_size = chunk_size
        self.overlap = overlap

    def get_chunks(self, filename: str) -> List[Fact]:
        _logger.info(f"Reading PDF file: {filename}")
        with pymupdf.open(filename) as doc:
            return [
                Fact(
                    text=page.get_text(),
                    metadata={"filename": filename, "page_number": i},
                )
                for i, page in enumerate(doc)
            ]



import asyncio
import os
import yaml

from unittest import TestCase

from wafl.config import Configuration
from wafl.dataclasses.dataclasses import Query
from wafl.knowledge.indexing_implementation import add_to_index, load_knowledge

_path = os.path.dirname(__file__)


class TestIndexing(TestCase):
    def test__path_can_be_added_to_index(self):
        data = _load_index()
        prior_count = len(data["paths"])
        add_to_index("files_to_index2")

        data = _load_index()
        current_count = len(data["paths"])
        self.assertEqual(current_count, prior_count + 1)

        data["paths"].remove("files_to_index2")
        with open("indices.yaml", "w") as file:
            file.write(yaml.dump(data))

    def test__indexed_files_can_be_retrieved(self):
        config = Configuration.load_local_config()
        knowledge = asyncio.run(load_knowledge(config))
        results = asyncio.run(
            knowledge.ask_for_facts(Query.create_from_text("How do I start WAFL"))
        )
        expected = "WAFL"
        self.assertIn(expected, results[0].text)

    def test__pdf_can_be_read(self):
        config = Configuration.load_local_config()
        knowledge = asyncio.run(load_knowledge(config))
        results = asyncio.run(
            knowledge.ask_for_facts(Query.create_from_text("What color is the sky?"))
        )
        expected = "green"
        self.assertIn(expected, results[0].text)


def _load_index():
    with open("indices.yaml", "r") as file:
        return yaml.safe_load(file.read())



import json
from datetime import datetime
from wafl.exceptions import CloseConversation

_db_filename = "db.json"


def close_conversation():
    raise CloseConversation()


def get_time():
    return datetime.now().strftime("%H,%M")


def get_date():
    return datetime.now().strftime("%Y-%m-%d")


def get_day():
    return datetime.now().strftime("%A")


def add_to_shopping_list(list_of_items_to_add):
    db = json.load(open(_db_filename))
    for item in list_of_items_to_add:
        if item not in db["shopping_list"]:
            db["shopping_list"].append(item)

    json.dump(db, open(_db_filename, "w"))

    return "Item added"


def remove_from_shopping_list(list_of_items_to_remove):
    db = json.load(open(_db_filename))
    for item in list_of_items_to_remove:
        if item in db["shopping_list"]:
            db["shopping_list"].remove(item)

    json.dump(db, open(_db_filename, "w"))

    return "Item removed"


def get_shopping_list():
    db = json.load(open(_db_filename))
    if db["shopping_list"] == []:
        return "nothing"

    return ", ".join(db["shopping_list"])



from wafl.readers.pdf_reader import PdfReader
from wafl.readers.text_reader import TextReader


class ReaderFactory:
    _chunk_size = 10000
    _overlap = 100
    _extension_to_reader_dict = {".pdf": PdfReader, ".txt": TextReader}

    @staticmethod
    def get_reader(filename):
        for extension, reader in ReaderFactory._extension_to_reader_dict.items():
            if extension in filename.lower():
                return reader(ReaderFactory._chunk_size, ReaderFactory._overlap)

        ### add pdf reader
        ### add metadata and show in the UI

        return TextReader(ReaderFactory._chunk_size, ReaderFactory._overlap)



from dataclasses import dataclass
from typing import List


@dataclass
class Rule:
    effect: "Fact"
    causes: List["Fact"]
    _max_indentation = 3
    indent_str = "  "

    def toJSON(self):
        return str(self)

    def get_string_using_template(self, effect_template: str) -> str:
        rule_str = effect_template.replace("{effect}", self.effect.text)
        rule_str = rule_str.replace("{clauses}", self._get_clauses())
        return rule_str

    def __str__(self):
        rule_str = self.effect.text + "\n"
        return rule_str + self._get_clauses()

    def _get_clauses(self) -> str:
        rule_str = ""
        for cause in self.causes:
            try:
                rule_str += self._recursively_add_clauses(cause)

            except TypeError as e:
                print(f"Error in rule:'''\n{rule_str}'''")
                print("Perhaps the YAML file is not well formatted.")
                print()
                raise e

        return rule_str[:-1]

    def _recursively_add_clauses(self, query: str, depth: int = 1) -> str:
        indentation = self.indent_str * depth
        if type(query) == str:
            return f"{indentation}- {query}\n"

        if type(query.text) == str:
            return f"{indentation}- {query.text}\n"

        if depth > self._max_indentation:
            return ""

        clause = list(query.text.keys())[0]
        rules_text = f"{indentation}- {clause}\n"
        for clauses in query.text.values():
            for cause_index, clause in enumerate(clauses):
                rules_text += self._recursively_add_clauses(clause, depth + 1)

        return rules_text



import asyncio

import numpy as np
import pyaudio

from wafl.connectors.factories.speaker_connector_factory import SpeakerConnectorFactory
from wafl.speaker.base_speaker import BaseSpeaker
from wafl.speaker.utils import convert_numbers_to_words


class FairSeqSpeaker(BaseSpeaker):
    def __init__(self, config):
        self._connector = SpeakerConnectorFactory.get_connector(config)
        self._p = pyaudio.PyAudio()
        self._input_chunk_size = 1024
        self._output_chunk_size = 16384
        self._volume_threshold = (
            config.get_value("listener_model")["listener_volume_threshold"] / 5e3
        )

    async def speak(self, text):
        text = convert_numbers_to_words(text)
        prediction = await self._connector.predict(text)
        wav = prediction["wav"]
        rate = prediction["rate"]
        stream = self._p.open(
            format=pyaudio.paFloat32,
            channels=1,
            rate=rate,
            input=True,
            output=True,
        )
        stream.start_stream()
        await asyncio.sleep(0.1)
        for i in range(0, len(wav), self._output_chunk_size):
            inp = stream.read(self._input_chunk_size)
            if _rms(inp) > self._volume_threshold:
                break
            stream.write(wav[i : i + self._output_chunk_size])
        stream.stop_stream()
        stream.close()
        await asyncio.sleep(0.1)


def _rms(frame):
    data = np.frombuffer(frame, dtype=np.float32)
    return np.std(data) / len(data)



import asyncio

import yaml

from wafl.answerer.entailer import Entailer
from wafl.config import Configuration
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.dummy_interface import DummyInterface
from wafl.logger.local_file_logger import LocalFileLogger

_logger = LocalFileLogger()

COLOR_YELLOW = "\033[93m"
COLOR_GREEN = "\033[92m"
COLOR_END = "\033[0m"


def get_action_list_and_expeted_list_from_yaml(filename, action_name):
    actions = yaml.safe_load(open("actions.yaml"))
    if action_name not in actions:
        raise ValueError(f"Action {action_name} not found in actions.yaml")

    actions_list = [item["action"] for item in actions[action_name]]
    expected_list = [item["expected"] for item in actions[action_name]]
    return actions_list, expected_list


def predict_action(config, actions_list, expected_list):
    interface = DummyInterface(to_utter=actions_list.copy(), print_utterances=True)
    conversation_events = ConversationEvents(
        config=config,
        interface=interface,
        logger=_logger,
    )
    entailer = Entailer(config)
    for expected in expected_list:
        asyncio.run(conversation_events.process_next())
        last_utterance = interface.get_utterances_list()[-1]

        if not last_utterance:
            raise ValueError("The agent did not say anything.")

        if expected and not asyncio.run(
            entailer.left_entails_right(
                last_utterance,
                expected,
                "\n".join(interface.get_utterances_list()[:-1]),
            )
        ):
            del entailer, conversation_events, interface
            raise ValueError(
                f"The utterance '{last_utterance}' does not entail '{expected}'."
            )


def run_action(action_name):
    print(COLOR_GREEN + f"Running the action {action_name}\n" + COLOR_END)
    actions_list, expected_list = get_action_list_and_expeted_list_from_yaml(
        "actions.yaml", action_name
    )
    config = Configuration.load_local_config()
    num_retries = 10
    success = False
    for _ in range(num_retries):
        try:
            predict_action(config, actions_list, expected_list)
            success = True
            break

        except (ValueError, SyntaxError) as e:
            print(COLOR_YELLOW + str(e) + COLOR_END)
            print(COLOR_GREEN + f"Retrying the action {action_name}..." + COLOR_END)

    if success:
        print(COLOR_GREEN + f"Action {action_name} finished." + COLOR_END)

    else:
        print(COLOR_YELLOW + f"Action {action_name} failed." + COLOR_END)


if __name__ == "__main__":
    run_action("action_1_summarise_guardian")



import os
from typing import List

from wafl.connectors.factories.llm_connector_factory import LLMConnectorFactory
from wafl.connectors.prompt_template import PromptTemplate

_path = os.path.dirname(__file__)


class LLMChatClient:
    def __init__(self, config):
        self._connector = LLMConnectorFactory.get_connector(config)
        self._config = config
        with open(self._config.get_value("prompt_filename")) as f:
            self.prompt = f.read()

    async def get_answer(self, text: str, dialogue: str, rules_text: List[str]) -> str:
        prompt = await self._get_answer_prompt(text, dialogue, "\n".join(rules_text))
        return await self._connector.generate(prompt)

    async def _get_answer_prompt(
        self, text: str, dialogue: "Conversation" = None, rules_text: str = None
    ) -> PromptTemplate:
        return PromptTemplate(
            system_prompt=self._get_system_prompt(text, rules_text),
            conversation=dialogue,
        )

    def _get_system_prompt(self, text, rules_text):
        return (
            self.prompt.replace("{facts}", text.strip())
            .replace("{rules}", rules_text.strip())
            .strip()
        )



import re
import traceback

from typing import List, Tuple

from wafl.exceptions import CloseConversation
from wafl.dataclasses.facts import Fact
from wafl.interface.conversation import Conversation, Utterance


def is_executable(text: str) -> bool:
    return "<execute>" in text


def create_one_liner(query_text):
    return Conversation(
        [
            Utterance(
                query_text,
                "user",
            )
        ]
    )


async def substitute_memory_in_answer_and_get_memories_if_present(
    answer_text: str,
) -> Tuple[str, List[str]]:
    matches = re.finditer(
        r"<remember>(.*?)</remember>|<remember>(.*?)$",
        answer_text,
        re.DOTALL | re.MULTILINE,
    )
    memories = []
    for match in matches:
        to_substitute = match.group(1)
        if not to_substitute:
            continue
        answer_text = answer_text.replace(match.group(0), "[Output in memory]")
        memories.append(to_substitute)

    answer_text = answer_text.replace("<br>", "\n")
    matches = re.finditer(r"<remember>(.*?)$", answer_text, re.DOTALL | re.MULTILINE)
    memories = []
    for match in matches:
        to_substitute = match.group(1)
        if not to_substitute:
            continue
        answer_text = answer_text.replace(match.group(0), "[Output in memory]")
        memories.append(to_substitute)

    return answer_text, memories


async def execute_results_in_answer(answer_text: str, module, functions) -> str:
    matches = re.finditer(
        r"<execute>(.*?)</execute>|<execute>(.*?\))$",
        answer_text,
        re.DOTALL | re.MULTILINE,
    )
    for match in matches:
        to_execute = match.group(1)
        if not to_execute:
            continue
        result = await _run_code(to_execute, module, functions)
        answer_text = answer_text.replace(match.group(0), result)

    matches = re.finditer(r"<execute>(.*?\))$", answer_text, re.DOTALL | re.MULTILINE)
    for match in matches:
        to_execute = match.group(1)
        if not to_execute:
            continue
        result = await _run_code(to_execute, module, functions)
        answer_text = answer_text.replace(match.group(0), result)

    return answer_text


async def _run_code(to_execute: str, module, functions) -> str:
    result = None
    for _ in range(3):
        try:
            if any(item + "(" in to_execute for item in functions):
                result = eval(f"module.{to_execute}")
                break

            else:
                ldict = {}
                exec(to_execute, globals(), ldict)
                if "result" in ldict:
                    result = str(ldict["result"])
                    break

        except NameError as e:
            match = re.search(r"\'(\w+)\' is not defined", str(e))
            if match:
                to_import = match.group(1)
                to_execute = f"import {to_import}\n{to_execute}"

        except CloseConversation as e:
            raise e

        except Exception as e:
            result = (
                f"Error while executing\n\n```python\n{to_execute}\n```\n\n{str(e)}"
            )
            traceback.print_exc()
            break

    if not result:
        result = f"\n```python\n{to_execute}\n```"

    return result


def get_text_from_facts_and_thresholds(
    facts_and_thresholds: List[Tuple[Fact, float]], memory: str
) -> List[str]:
    text_list = []
    for item in facts_and_thresholds:
        if item[0].text not in memory:
            text = item[0].text
            if item[0].metadata:
                text = (
                    f"Metadata for the following text: {str(item[0].metadata)}"
                    + "\n"
                    + text
                )
            text_list.append(text)

    return text_list


def add_dummy_utterances_to_continue_generation(
    conversation: Conversation, answer_text: str
):
    conversation.add_utterance(
        Utterance(
            answer_text,
            "bot",
        )
    )
    conversation.add_utterance(
        Utterance(
            "Continue",
            "user",
        )
    )


def add_memories_to_facts(facts: str, memories: List[str]) -> str:
    return facts + "\n" + "\n".join(memories)



import asyncio

from unittest import TestCase

from wafl.config import Configuration
from wafl.events.conversation_events import ConversationEvents
from wafl.interface.dummy_interface import DummyInterface
from wafl.parsing.rules_parser import get_facts_and_rules_from_text

wafl_example = """
rules:
  - the user says "my name is bob":
    - You must answer the user by writing "the horse is tall"
    
  - the user says their name:
    - reply casually to the conversation"
    
  - the user wants to buy coffee:
    - the bot says the coffee prices:
        - decaf is 1.50
        - regular is 1.00
        - espresso is 2.00
    - ask for which price range
    - tell them the right coffee
"""


class TestRules(TestCase):
    def test__rules_can_be_triggered(self):
        interface = DummyInterface(
            to_utter=[
                "my name is bob",
            ]
        )
        config = Configuration.load_local_config()
        config.set_value("rules", wafl_example)
        conversation_events = ConversationEvents(
            config=config,
            interface=interface,
        )
        asyncio.run(conversation_events.process_next())
        expected = "the horse is tall"
        self.assertIn(expected, interface.get_utterances_list()[-1].lower())

    def test__rules_are_not_always_triggered(self):
        interface = DummyInterface(
            to_utter=[
                "my name is Frank",
            ]
        )
        config = Configuration.load_local_config()
        config.set_value("rules", wafl_example)
        conversation_events = ConversationEvents(
            config=config,
            interface=interface,
        )
        asyncio.run(conversation_events.process_next())
        unexpected = "bot: the horse is tall"
        self.assertNotEqual(unexpected, interface.get_utterances_list()[-1])

    def test__rules_can_nest(self):
        interface = DummyInterface(
            to_utter=[
                "I want to buy coffee",
            ]
        )
        config = Configuration.load_local_config()
        config.set_value("rules", wafl_example)
        conversation_events = ConversationEvents(
            config=config,
            interface=interface,
        )
        asyncio.run(conversation_events.process_next())
        self.assertIn("decaf", interface.get_facts_and_timestamp()[0][1])
        self.assertIn("regular", interface.get_facts_and_timestamp()[0][1])
        self.assertIn("espresso", interface.get_facts_and_timestamp()[0][1])

    def test__nested_rules_are_printed_correctly(self):
        rule_text = """
rules:        
    - the user wants to know the time:
        - output "The time is <execute>get_time()</execute>":
            - if the time is before 12:00 say "Good morning"
            - if the time is after 12:00 say "Good afternoon"        
        """.strip()

        facts_and_rules = get_facts_and_rules_from_text(rule_text)
        rule = facts_and_rules["rules"][0]
        expected = """
the user wants to know the time
  - output "The time is <execute>get_time()</execute>"
    - if the time is before 12:00 say "Good morning"
    - if the time is after 12:00 say "Good afternoon"        
        """.strip()
        self.assertEqual(expected, str(rule).strip())
